import std.stdio;
import std.string;
import std.regexp;
import std.format;
import std.file;
import std.thread;
import std.process;

import SDL;
import SDL_ttf;
import SDL_image;

import http;

version (Win32) {
    import std.c.windows.windows;
    extern(Windows) export HGLOBAL GlobalAlloc(
        uint uFlags,
        uint dwBytes);
    enum : uint { GMEM_DDESHARE = 0x2000 }
    enum : uint { GMEM_MOVEABLE = 0x0002 }
    extern(Windows) export LPVOID GlobalLock(
        HGLOBAL hMem);
    extern(Windows) export LPSTR lstrcpyA(
        LPSTR lpString1,
        LPCSTR lpString2);
    extern(Windows) export LPWSTR lstrcpyW(
        LPWSTR lpString1,
        LPCWSTR lpString2);

    extern(Windows) export BOOL OpenClipboard(
        void* hWndNewOwner);
    extern(Windows) export BOOL CloseClipboard();
    extern(Windows) export BOOL EmptyClipboard();
    extern(Windows) export HANDLE SetClipboardData(
        uint uFormat,
        HANDLE hMem);
    enum : uint { CF_TEXT = 1 }
}

SDL_Surface* mem2surf(char[] mem) {
    SDL_RWops* rwops = SDL_RWFromMem(&mem[0], mem.length);
    SDL_Surface* s = IMG_LoadGIF_RW(rwops);
    return s;
}

class Image {
    char[] title;
    char[] path;
    SDL_Surface* surf;
    char[] gif;
}

class Serif : Thread {
    this() {
        h = new Http11("serif.hatelabo.jp", 80);
        h.close = true;
        speed = 10000;
    }

    override int run() {
        Uint32 prevTicks;
        while (!done) {
            try {
                Uint32 ticks = SDL_GetTicks();
                if (!stop && (!cur || prevTicks + speed < ticks)) {
                    Image c = search();
                    Image p = cur;
                    cur = c;
                    if (p) {
                        SDL_FreeSurface(p.surf);
                        delete p.gif;
                    }
                    prevTicks = SDL_GetTicks();
                }
            }
            catch (Object o) {
                writefln(o);
            }
            SDL_Delay(100);
        }
        return 0;
    }

    Image get() {
        return cur;
    }

    char[] http_get(char[] p) {
        writefln(p);
        loading = SDL_GetTicks();
        HttpResponse res = h.get(p);
        loading = 0;
        return res.messageBody;
    }

    Image search() {
        if (error && error + 30000 > SDL_GetTicks()) return null;

        if (images.length == 0) {
            page++;
            char[] p = "/seriflist/?page=" ~ std.string.toString(page);
            char[] c = http_get(p);

            foreach (m; RegExp("<img alt=\"([^\"]+) - Hatena Serif\" src=\"http://serif.hatelabo.jp(/images/cache/[/a-zA-Z0-9]+\\.gif)").search(c)) {
                Image im = new Image;
                im.title = m.match(1);
                im.path = m.match(2);
                images ~= im;
            }
            if (images.length == 0) {
                page--;
                error = SDL_GetTicks();
                return null;
            }
        }

        Image im = images[0];
        images = images[1..$];
        char[] c = http_get(im.path);
        if (!c || c.length == 0) return null;
        char[] fix = c[1..$];
        im.surf = mem2surf(fix);
        im.gif = fix;
        return im;
    }

    void end() {
        done = true;
    }

    void flipStop() {
        stop = !stop;
    }

    void speedUp() {
        speed -= 1000;
        if (speed < 5000) speed = 5000;
        setMsg(std.string.toString(speed/1000) ~ " Sec/Pic");
    }
    void speedDown() {
        speed += 1000;
        setMsg(std.string.toString(speed/1000) ~ " Sec/Pic");
    }

    void setMsg(char[] m) {
        msg = m;
        msgTicks = SDL_GetTicks();
    }

    char[] status() {
        if (msgTicks + 500 > SDL_GetTicks()) {
            return msg;
        }
        else if (stop) return "PAUSE";
        else if (error) return "ERROR";
        else if (loading) {
            Uint32 lt = SDL_GetTicks() - loading;
            return "LOADING" ~ repeat(".", (lt/100)%10);
        }
        else return null;
    }

private:
    Http11 h;
    int page;
    Image[] images;
    Image cur;
    bool done;
    Uint32 loading;
    Uint32 error;
    bool stop;
    char[] msg;
    Uint32 msgTicks;
    int speed;
}

void render(TTF_Font* f, SDL_Surface* s, char[] txt, int x, int y,
            int r = 255, int g = 255, int b = 255) {
    SDL_Rect dr;
    dr.x = x;
    dr.y = y;
    SDL_Color fg;
    fg.r = r;
    fg.g = g;
    fg.b = b;
    SDL_Surface* t = TTF_RenderUTF8_Blended(f, toStringz(txt), fg);
    SDL_BlitSurface(t, null, s, &dr);
    SDL_FreeSurface(t);
}

extern (C) {
    void* fontlist_open();
    void fontlist_close(void* fl);
    char* fontlist_next_file(void* fl);
}
char[] getFontname() {
    version (Win32) {
        void* fl = fontlist_open();
        char* fp;
        char[][] fonts;
        while ((fp = fontlist_next_file(fl)) != null) {
            char[] f = std.string.toString(fp).dup;
            if (std.string.find(f, "msmincho") != -1) {
                fontlist_close(fl);
                return f;
            }
            if (std.string.find(f, "msgothic") != -1) {
                if (fonts.length == 0) {
                    fonts ~= f;
                }
                else {
                    fonts[0] = f;
                }
            }
            else {
                fonts ~= f;
            }
        }
        if (fonts.length == 0) return null;
        fontlist_close(fl);
        writefln("could not find msmincho");
        return fonts[0];
    }
    else {
        return "ipamp-mona.ttf";
    }
}

void save(Image im) {
    version (Win32) {
        char[] file = im.title ~ ".gif";
    }
    else {
        char[] file = im.path;
        file = file[std.string.rfind(file, "/")+1..$];
    }
    std.file.write("save/" ~ file, im.gif);
}

void clip(char[] p) {
    if (p.length < 20) return;
    char[] url = "http://serif.hatelabo.jp/" ~ p[14..$-4] ~ "/";
    version (Win32) {
        // http://nienie.com/~masapico/api_SetClipboardData.html
        HGLOBAL hText;
        hText = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE, url.length+1);
        if (!hText) {
            writefln("GlobalAlloc failed");
            return;
        }
        char* pText = cast(char*)GlobalLock(hText);
        if (!pText) {
            writefln("GlobalLock failed");
            return;
        }
        writefln("copying " ~ url);
        lstrcpyA(pText, toStringz(url));
        GlobalUnlock(hText);

        if (OpenClipboard(null) == FALSE) {
            writefln("OpenClipboard failed");
        }
        EmptyClipboard();
        SetClipboardData(CF_TEXT, hText);
        CloseClipboard();
    }
    else {
        char[] cmd = "echo -n " ~ url ~ " | xclip";
        system(cmd);
    }
}

int main(char[][] args) {
    SDL_Init(SDL_INIT_VIDEO);
    TTF_Init();

    SDL_Surface* dpy = SDL_SetVideoMode(800, 600, 32, 0);
    SDL_Surface* tmp = SDL_CreateRGBSurface(0, 800, 600, 32, 0,0,0,0);
    SDL_Surface* scr = SDL_DisplayFormat(tmp);
    SDL_FreeSurface(tmp);
    SDL_SetAlpha(scr, SDL_SRCALPHA|SDL_RLEACCEL, 30);
    SDL_WM_SetCaption("paseri", null);

    char[] fontname = getFontname();
    if (!fontname) {
        writefln("font not found");
        return 1;
    }
    version (Win32) {
    }
    else {
        if (!exists(fontname)) {
            writefln(fontname ~ ": font not found (" ~
                     std.string.toString(SDL_GetError()) ~ ")");
            return 1;
        }
    }
    writefln(fontname ~ ": loading font");
    TTF_Font* font = TTF_OpenFont(toStringz(fontname), 30);
    if (!font) {
        writefln(fontname ~ ": load failed");
        return 1;
    }

    Serif serif = new Serif();
    serif.start();

    bool full = false;

    bool done = false;
    while (!done) {
        Image im = serif.get();
        try {
            if (im) {
                SDL_Surface* s = im.surf;
                SDL_FillRect(scr, null, SDL_MapRGB(scr.format, 0,0,0));
                SDL_Rect dr;
                dr.x = (scr.w - s.w) / 2;
                dr.y = (scr.h - s.h) / 2;
                SDL_BlitSurface(s, null, scr, &dr);

                render(font, scr, im.title, 15, 15);
            }

            SDL_BlitSurface(scr, null, dpy, null);

            char[] status = serif.status;
            if (status) {
                render(font, dpy, status, scr.w-200, scr.h-45, 255,0,0);
            }
        }
        catch (Object o) {
            writefln(o);
        }

        SDL_Flip(dpy);

        SDL_Delay(100);

        SDL_Event ev;
        while (SDL_PollEvent(&ev)) {
            switch (ev.type) {
            case SDL_QUIT:
                done = true;
                break;
            case SDL_KEYDOWN:
                SDLKey k = ev.key.keysym.sym;
                if (k == 'q' || k == 'Q' || k == SDLK_ESCAPE) {
                    done = true;
                }
                else if (k == 'c') {
                    if (im) {
                        clip(im.path);
                        serif.setMsg("Clipped URL");
                    }
                }
                else if (k == 'p') {
                    serif.flipStop();
                }
                else if (k == SDLK_RIGHT) {
                    serif.speedUp();
                }
                else if (k == SDLK_LEFT) {
                    serif.speedDown();
                }
                else if (k == 's') {
                    if (im) {
                        save(im);
                        serif.setMsg("Saved");
                    }
                }
                else if (k == 'f') {
                    full = !full;
                    Uint32 flags = 0;
                    if (full) flags |= SDL_FULLSCREEN;
                    dpy = SDL_SetVideoMode(800, 600, 32, flags);
                }
                break;
            default:
            }
        }
    }

    serif.end();
    serif.wait();

    TTF_Quit();
    SDL_Quit();

    return 0;
}
